import pyarbtools
import matplotlib.pyplot as plt
import numpy as np

def new_mod_test():
    pyarbtools.wfmBuilder.digmod_generator(modType='apsk64')


def vsg_dig_mod_example(vsgIpAddress, vsaIpAddress):
    """Generates and plays 1 MHz 16 QAM signal with 0.35 alpha RRC filter
    @ 1 GHz CF with a generic VSG."""

    # Create VSG object
    vsg = pyarbtools.instruments.VSG(vsgIpAddress, port=5025, timeout=15, reset=True)

    # Signal generator configuration variables
    amplitude = -5
    fs = 113e6
    freq = 1e9

    # Configure signal generator
    vsg.configure(amp=amplitude, fs=fs, cf=freq)
    vsg.sanity_check()
    vsg.err_check()

    # Waveform creation variables
    symRate = 19e6
    modType = 'qam64'
    psFilter = 'rootraisedcosine'
    alpha = 0.35
    name = 'digmod'

    print('Creating waveform.')
    # This is the new digital modulation waveform creation function
    data = pyarbtools.wfmBuilder.digmod_generator(fs=fs, symRate=symRate, modType=modType, filt=psFilter, numSymbols=1000,
                                                  alpha=alpha)

    # plt.plot(np.concatenate([data.real[-100:], data.real[:100]]))
    # plt.plot(np.concatenate([data.imag[-100:], data.imag[:100]]))
    # plt.show()

    # Download and play waveform
    vsg.download_wfm(data, wfmID=name)
    vsg.play(name)

    print('Setting up VSA.')
    # Create VSA object
    vsa = pyarbtools.vsaControl.VSA(vsaIpAddress, vsaHardware='Morgan Office PXA', timeout=30, reset=False)

    # Select a digital demod measurement and configure it to measure the saved waveform
    vsa.set_measurement('ddemod')
    if psFilter.lower() == 'rootraisedcosine':
        mFilter = 'rootraisedcosine'
        rFilter = 'raisedcosine'
    elif psFilter.lower() == 'raisedcosine':
        mFilter = 'none'
        rFilter = 'raisedcosine'
    else:
        raise Exception('Invalid filter type chosen.')

    # Configure digital demodulation in VSA
    vsa.configure_ddemod(cf=freq, amp=0, span=symRate * 2, modType=modType, symRate=symRate, measFilter=mFilter, refFilter=rFilter, filterAlpha=alpha,
                         measLength=4096, eqState=False)

    # Perform a single-shot replay in VSA
    # vsa.acquire_single()
    vsa.acquire_continuous()

    # Check for errors and gracefully disconnect
    vsa.err_check()
    vsa.disconnect()
    # Check for erros and gracefully disconnect
    vsg.err_check()
    vsg.disconnect()




def wfm_to_vsa_example(ipAddress):
    """This function creates a "perfect" digitally modulated waveform, exports it to a csv file,
    recalls it into VSA, and configures VSA to analyze it."""

    # Waveform creation variables
    symRate = 11e6
    fs = 200e6
    modType = 'qam64'
    psFilter = 'rootraisedcosine'
    alpha = 0.35
    fileName = 'C:\\Temp\\wfm.csv'
    fileFormat = 'csv'

    print('Creating waveform.')
    # This is the new digital modulation waveform creation function
    data = pyarbtools.wfmBuilder.digmod_generator(fs=fs, symRate=symRate, modType=modType, filt=psFilter, numSymbols=1000, alpha=alpha)

    plt.plot(np.concatenate([data.real[-100:], data.real[:100]]))
    plt.plot(np.concatenate([data.imag[-100:], data.imag[:100]]))
    plt.show()


    print('Exporting waveform.')
    # Export the waveform to a csv file
    pyarbtools.wfmBuilder.export_wfm(data, fileName, True, fs)

    print('Setting up VSA.')
    # Create VSA object
    vsa = pyarbtools.vsaControl.VSA(ipAddress, vsaHardware=None, timeout=30, reset=False)

    # Select a digital demod measurement and configure it to measure the saved waveform
    vsa.set_measurement('ddemod')
    if psFilter.lower() == 'rootraisedcosine':
        mFilter = 'rootraisedcosine'
        rFilter = 'raisedcosine'
    elif psFilter.lower() == 'raisedcosine':
        mFilter = 'none'
        rFilter = 'raisedcosine'
    else:
        raise Exception('Invalid filter type chosen.')

    # Configure digital demodulation in VSA
    vsa.configure_ddemod(amp=0, modType=modType, symRate=symRate, measFilter=mFilter, refFilter=rFilter, filterAlpha=alpha,
                         measLength=1000, eqState=False)

    # Recall csv file we exported earlier
    vsa.recall_recording(fileName, fileFormat=fileFormat)

    # Perform a single-shot replay in VSA
    vsa.acquire_single()

    # Check for errors and gracefully disconnect
    vsa.err_check()
    vsa.disconnect()

def ofdm_to_vsa_example(ipAddress):
    """This function creates a "perfect" digitally modulated waveform, exports it to a csv file,
    recalls it into VSA, and configures VSA to analyze it."""

    # Waveform creation variables
    fftLength = 8
    fs = fftLength * 16
    print(fs)
    modType = 'qpsk'
    psFilter = 'rootraisedcosine'
    alpha = 0.35
    fileName = 'C:\\Temp\\ofdm.csv'
    fileFormat = 'csv'

    print('Creating waveform.')
    # This is the new digital modulation waveform creation function
    data, resourceMap, modMap, pilotValues = pyarbtools.wfmBuilder.ofdm(fs=fs, fftLength=fftLength, modType=modType, numSymbols=1)

    print('Exporting waveform.')
    # Export the waveform to a csv file
    pyarbtools.wfmBuilder.export_wfm(data, fileName, True, fs)

    print('Setting up VSA.')
    # Create VSA object
    vsa = pyarbtools.vsaControl.VSA(ipAddress, vsaHardware=None, timeout=20, reset=False)

    # Select a digital demod measurement and configure it to measure the saved waveform
    # vsa.set_measurement('ddemod')
    # if psFilter.lower() == 'rootraisedcosine':
    #     mFilter = 'rootraisedcosine'
    #     rFilter = 'raisedcosine'
    # elif psFilter.lower() == 'raisedcosine':
    #     mFilter = 'none'
    #     rFilter = 'raisedcosine'
    # else:
    #     raise Exception('Invalid filter type chosen.')

    # Configure digital demodulation in VSA
    # vsa.configure_ddemod(amp=0, modType=modType, symRate=symRate, measFilter=mFilter, refFilter=rFilter, filterAlpha=alpha,
    #                      measLength=1000, eqState=False)

    vsa.write('measure:nselect 1')
    vsa.write('measure:configure customofdm')
    vsa.query('*opc?')

    vsa.configure_customofdm(fs=fs, fftLength=fftLength, guardLowerSubcarriers=0, guardUpperSubcarriers=0, modType=modType)

    # format rmap to a format VSA can understand, i.e. a string with comma-separated values.
    print(resourceMap)
    # vsaResourceMap = ','.join([str(index) for index in resourceMap])
    vsa.write(f'sense:customofdm:format:rmap {resourceMap}')

    print(modMap)
    # vsaModMap = ','.join([str(index) for index in modMap])
    vsa.write(f'sense:customofdm:format:rmodulation {modMap}')

    vsa.write(f'sense:customofdm:format:iq:pilot {pilotValues}')

    # Recall csv file we exported earlier
    vsa.recall_recording(fileName, fileFormat=fileFormat)

    # Perform a single-shot replay in VSA
    vsa.acquire_single()

    # Check for errors and gracefully disconnect
    vsa.err_check()
    vsa.disconnect()



def wraparound_experimentation():
    # Waveform creation variables
    symRate = 10e6
    fs = 200e6
    modType = 'qam64'
    psFilter = 'rootraisedcosine'
    alpha = 0.35
    fileName = 'C:\\Temp\\wfm.csv'
    fileFormat = 'csv'

    print('Creating waveform.')

    for x in range(1,2):

        # This is the new digital modulation waveform creation function
        data = pyarbtools.wfmBuilder.digmod_generator(fs=fs, symRate=symRate, modType=modType, filt=psFilter, numSymbols=100,
                                                      alpha=alpha)

        i = data.real
        q = data.imag

        # show both the end and beginning of the waveform
        iPlot = np.concatenate([i[-500:], i[:500]])
        qPlot = np.concatenate([q[-500:], q[:500]])
        plt.axvline(500, c='r')

        # plt.subplot(3, 3, x)
        plt.plot(iPlot)
        # plt.subplot(3, 3, x)
        plt.plot(qPlot)
    plt.show()

        # print('Exporting waveform.')
        # Export the waveform to a csv file
        # pyarbtools.wfmBuilder.export_wfm(data, fileName, True, fs)

def gui_test():
    pyarbtools.gui.main()


def main():
    # ofdm_to_vsa_example('127.0.0.1')
    # wfm_to_vsa_example('127.0.0.1')
    # new_mod_test()
    # vsg_dig_mod_example('192.168.50.124', '127.0.0.1')
    # wraparound_experimentation()
    gui_test()

if __name__ == '__main__':
    main()
